from logging import getLogger
from numpy import mean, std
from pickle import dump
from wonambi import Dataset
from wonambi.trans import math, concatenate
from bidso import Task, Electrodes

lg = getLogger(__name__)


def read_ieeg_block(filename, electrode_file, cond, minimalduration, output_dir):
    d = Dataset(filename, bids=True)
    markers = d.read_markers()

    electrodes = Electrodes(electrode_file)
    elec_names = [x['name'] for x in electrodes.electrodes.tsv]
    elec_names = [x for x in elec_names if x in d.header['chan_name']]  # exclude elec location that have no corresponding channel

    clean_labels = _reject_channels(d, elec_names, cond, minimalduration)

    outputs = []
    for cond_name, cond_mrk in cond.items():
        block_beg = []
        block_end = []

        for mrk in markers:

            if mrk['name'] in cond_mrk:
                dur = (mrk['end'] - mrk['start'])
                if dur >= minimalduration:
                    block_beg.append(mrk['start'])
                    block_end.append(mrk['end'])

        data = d.read_data(begtime=block_beg, endtime=block_end, chan=clean_labels)

        output_task = Task(filename)
        output_task.extension = '.pkl'
        output_task.task += cond_name
        output_file = output_dir / output_task.get_filename()
        with output_file.open('wb') as f:
            dump(data, f)
        outputs.append(output_file)

    return outputs


def _reject_channels(d, elec_names, cond, minimalduration):
    markers = d.read_markers()
    block_beg = []
    block_end = []
    for mrk in markers:
        if mrk['name'] in cond.values():
            dur = (mrk['end'] - mrk['start'])
            if dur >= minimalduration:

                block_beg.append(mrk['start'])
                block_end.append(mrk['end'])

    data = d.read_data(chan=elec_names, begtime=block_beg, endtime=block_end)
    data = concatenate(data, 'time')

    clean_labels = reject_channels(data, 3)
    return clean_labels


def reject_channels(dat, reject_chan_thresh):
    dat_std = math(dat, operator_name='std', axis='time')
    THRESHOLD = reject_chan_thresh
    x = dat_std.data[0]
    thres = [mean(x) + THRESHOLD * std(x)]
    clean_labels = list(dat_std.chan[0][dat_std.data[0] < thres])
    return clean_labels



def read_prf():
    prestim = float(prestim)
    poststim = float(poststim)

    d = Dataset(ieeg_file, bids=True)
    events = array([x['start'] for x in d.read_markers()])

    data = d.read_data(
        begtime=list(events - prestim),
        endtime=list(events + poststim + 1 / d.header['s_freq']))
    data.attr['stimuli'] = read_prf_stimuli(d.dataset.task)
    all_data = (data, )
    conds = ['', ]



def read_prf_stimuli(task):
    """Read stimuli to compute the PRF

    Parameters
    ----------
    task : instance of bidso.Task
        task containing the events and filename
    """
    stimuli_dir = find_root(task.filename) / 'stimuli'

    stim_file = stimuli_dir / task.events.tsv[0]['stim_file']
    if stim_file.suffix == '.npy':
        stimuli = load(stim_file)

    elif stim_file.suffix == '.mat':
        mat = loadmat(stim_file)
        stimuli = mat['stimulus'][0, 0]['images']

    stim_file_index = array([int(x['stim_file_index']) - 1 for x in task.events.tsv])
    stimuli = stimuli[:, :, stim_file_index]

    return stimuli



from logging import getLogger
from subprocess import run

from bidso.find import find_in_bids

from ..utils import ENVIRON

lg = getLogger(__name__)


def main(analysis_dir, freesurfer_dir):
    """
    coreg feat with freesurfer

    Parameters
    ----------
    analysis_dir : path

    freesurfer_dir : path

    """

    for feat_path in find_in_bids(analysis_dir, generator=True, extension='.feat'):
        lg.debug(f'Reading {feat_path}')
        coreg_feat2freesurfer(feat_path, freesurfer_dir)


def coreg_feat2freesurfer(feat_file, FREESURFER_PATH):
    """This needs to be improved with object-oriented feat"""
    cmd = ['reg-feat2anat', '--feat', str(feat_file), '--subject', feat_file.name.split('_')[0]]
    run(cmd,
        env={**ENVIRON, 'SUBJECTS_DIR': str(FREESURFER_PATH)},
        cwd=str(feat_file.parent))



from pickle import load, dump
from logging import getLogger
from numpy import empty, array
from scipy.signal import welch

from wonambi.trans import timefrequency
from wonambi.datatype import ChanTimeFreq
from bidso.utils import replace_extension

lg = getLogger(__name__)

HALFBANDWIDTH = 2


def compute_powerspectrum(ieeg_file, method, taper, duration, output_dir):
    """
    compute psd

    Parameters
    ----------
    method : str
        "spectrogram" or "dh2012"
    taper : str
        "dpss", "boxcar", "hann" (only if method=='spectrogram')
    duration : float
        duration of the trials
    """

    with ieeg_file.open('rb') as f:
        dat = load(f)

    if method == 'spectrogram':
        freq = compute_frequency(dat, taper, duration)
    elif method == 'dh2012':
        freq = compute_welch_dh2012(dat, duration)
    else:
        raise ValueError(f'Unknown method "{method}"')

    output_file = output_dir / replace_extension(ieeg_file.name, 'psd.pkl')
    with output_file.open('wb') as f:
        dump(freq, f)

    return output_file


def compute_frequency(dat, taper, duration):
    """Remove epochs which have very high activity in high-freq range, then
    average over time (only high-freq range) and ALL the frequencies."""

    dat = timefrequency(
        dat,
        method='spectrogram',
        taper=taper,
        duration=duration,
        halfbandwidth=HALFBANDWIDTH,
        )

    return dat


def compute_welch_dh2012(data, duration):
    NPERSEG = 102
    NFFT = data.s_freq * duration

    freq = ChanTimeFreq()
    freq.s_freq = data.s_freq
    freq.start_time = data.start_time
    freq.axis['chan'] = data.axis['chan']
    freq.axis['freq'] = empty(data.number_of('trial'), dtype='O')
    freq.axis['time'] = empty(data.number_of('trial'), dtype='O')
    freq.data = empty(data.number_of('trial'), dtype='O')

    for i, x in enumerate(data.data):
        [f, Pxx] = welch(x, window='hamming', fs=data.s_freq, nperseg=NPERSEG,
                         nfft=NFFT, noverlap=0, detrend=False)
        freq.freq[i] = f
        freq.time[i] = array([i, ], dtype='float')
        freq.data[i] = Pxx[:, None, :]

    return freq



from hashlib import md5
import gzip


def compute_md5(p):
    """Compute m5sum for a file. If the file is .gz (in the case of .nii.gz)
    then it reads the archived version (the .gz contains metadata that changes
    every time)
    """
    md5_ = md5()

    if p.suffix == '.gz':
        f = gzip.open(p, 'rb')
        val = f.read()

    elif p.suffix == '.vhdr':
        # skip first two lines because it contains the time the file was written
        f = f.open('rb')
        val = b'\n'.join(f.read().split(b'\n')[2:])

    else:
        f = p.open('rb')
        val = f.read()

    md5_.update(val)
    f.close()

    return md5_.hexdigest()



from logging import getLogger
from pathlib import Path
from subprocess import run, PIPE
from tempfile import TemporaryDirectory

from nibabel import Nifti1Image
from nibabel import load as niload
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage.morphology import binary_closing

from ..utils import check_subprocess


GAUSSIAN_FILTER = 1
CLOSING_ITER = 15
SMOOTH_ITERATION = 60

lg = getLogger(__name__)


def fill_surface(surf_in, surf_smooth):

    with TemporaryDirectory() as tmpdir:
        lg.debug(f'Temporary Directory for fill_surface: {tmpdir}')
        tmpdir = Path(tmpdir)

        vol_file = tmpdir / 'vol.mgz'
        vol_filled = tmpdir / 'filled.nii.gz'
        surf_filled = tmpdir / 'filled.surf'

        p = run([
            'mris_fill',
            '-c', '-r', '1',
            str(surf_in),
            str(vol_file),
            ], stdout=PIPE, stderr=PIPE)
        check_subprocess(p)

        _close_volume(vol_file, vol_filled)

        p = run([
            'mri_tessellate',
            str(vol_filled),
            '1',
            str(surf_filled),
            ], stdout=PIPE, stderr=PIPE)
        check_subprocess(p)

        p = run([
            'mris_smooth',
            '-nw',
            '-n', str(SMOOTH_ITERATION),
            str(surf_filled),
            str(surf_smooth)
            ], stdout=PIPE, stderr=PIPE)
        check_subprocess(p)


def _close_volume(vol_file, filled):

    vol = niload(str(vol_file))
    volume = vol.get_data()
    v = gaussian_filter(volume, GAUSSIAN_FILTER)

    # binarize
    v[v <= 25 / 255] = 0
    v[v > 25 / 255] = 1

    closed = binary_closing(v, iterations=CLOSING_ITER)
    n = Nifti1Image(closed.astype('float32'), vol.affine)
    n.to_filename(str(filled))

    return filled



from nipype import Function


def wrapper_corr(fmri_file, ecog_file, output_dir='./corr_values', pvalue=0.05):
    from pathlib import Path
    from boavus.corr.corrfmri import compute_corr_ecog_fmri

    return str(compute_corr_ecog_fmri(
        Path(fmri_file).resolve(),
        Path(ecog_file).resolve(),
        Path(output_dir).resolve(),
        pvalue))

def wrapper_summary(in_files, ecog_files, fmri_files, output_dir='./output'):
    from pathlib import Path
    from boavus.corr.summary import collect_corr

    collect_corr(in_files, ecog_files, fmri_files, Path(output_dir).resolve())


function_corr = Function(
    input_names=[
        'fmri_file',
        'ecog_file',
        'output_dir',
        'pvalue',
    ],
    output_names=[
        'out_file',
    ],
    function=wrapper_corr,
    )


function_corr_summary = Function(
    input_names=[
        'in_files',
        'ecog_files',
        'fmri_files',
    ],
    function=wrapper_summary,
    )



from nipype import Workflow, Node, MapNode
from nipype.interfaces.utility import IdentityInterface
from ..ieeg import (function_ieeg_read,
                    function_ieeg_preprocess,
                    function_ieeg_powerspectrum,
                    function_ieeg_compare,
                    )


def workflow_ieeg(PARAMETERS):

    input = Node(IdentityInterface(fields=['ieeg', 'electrodes']), name='input')

    node_read = Node(function_ieeg_read, name='read')
    node_read.inputs.conditions = PARAMETERS['read']['conditions']
    node_read.inputs.minimalduration = PARAMETERS['read']['minimalduration']

    node_preprocess = MapNode(function_ieeg_preprocess, name='preprocess', iterfield=['ieeg', ])
    node_preprocess.inputs.duration = PARAMETERS['preprocess']['duration']
    node_preprocess.inputs.reref = PARAMETERS['preprocess']['reref']
    node_preprocess.inputs.offset = PARAMETERS['preprocess']['offset']

    node_frequency = MapNode(function_ieeg_powerspectrum, name='powerspectrum', iterfield=['ieeg', ])
    node_frequency.inputs.method = PARAMETERS['powerspectrum']['method']
    node_frequency.inputs.taper = PARAMETERS['powerspectrum']['taper']
    node_frequency.inputs.duration = PARAMETERS['powerspectrum']['duration']

    node_compare = Node(function_ieeg_compare, name='ecog_compare')
    node_compare.inputs.frequency = PARAMETERS['ecog_compare']['frequency']
    node_compare.inputs.baseline = PARAMETERS['ecog_compare']['baseline']
    node_compare.inputs.method = PARAMETERS['ecog_compare']['method']
    node_compare.inputs.measure = PARAMETERS['ecog_compare']['measure']

    w = Workflow('ieeg')

    w.connect(input, 'ieeg', node_read, 'ieeg')
    w.connect(input, 'electrodes', node_read, 'electrodes')
    w.connect(node_read, 'ieeg', node_preprocess, 'ieeg')
    w.connect(node_preprocess, 'ieeg', node_frequency, 'ieeg')
    w.connect(node_frequency, 'ieeg', node_compare, 'in_files')

    return w



from nipype import Workflow, Node
from nipype.interfaces.fsl import FEAT, BET, FLIRT, Threshold
from nipype.interfaces.freesurfer import ReconAll
from nipype.interfaces.utility import IdentityInterface
from numpy import arange

from ..fsl import function_prepare_design
from ..fmri import (function_fmri_compare,
                    function_fmri_atelec,
                    function_fmri_graymatter,
                    )

UPSAMPLE_RESOLUTION = 1
DOWNSAMPLE_RESOLUTION = 4
GRAYMATTER_THRESHOLD = 0.2


def workflow_fmri(PARAMETERS, FREESURFER_PATH):
    """TODO: input and output"""

    input = Node(IdentityInterface(fields=['subject', 'T1w', 'bold', 'electrodes']), name='input')

    node_bet = Node(BET(), name='bet')
    node_bet.inputs.frac = 0.5
    node_bet.inputs.vertical_gradient = 0
    node_bet.inputs.robust = True

    node_featdesign = Node(function_prepare_design, name='feat_design')

    node_feat = Node(FEAT(), name='feat')

    node_compare = Node(function_fmri_compare, name='fmri_compare')
    node_compare.inputs.measure = PARAMETERS['fmri_compare']['measure']
    node_compare.inputs.normalize_to_mean = PARAMETERS['fmri_compare']['normalize_to_mean']

    node_upsample = Node(FLIRT(), name='upsample')  # not perfect, there is a small offset
    node_upsample.inputs.apply_isoxfm = UPSAMPLE_RESOLUTION
    node_upsample.inputs.interp = 'nearestneighbour'

    node_downsample = Node(FLIRT(), name='downsample')  # not perfect, there is a small offset
    node_downsample.inputs.apply_xfm = True
    node_downsample.inputs.uses_qform = True
    # node_downsample.inputs.apply_isoxfm = DOWNSAMPLE_RESOLUTION
    node_downsample.inputs.interp = 'nearestneighbour'

    node_threshold = Node(Threshold(), name='threshold')
    node_threshold.inputs.thresh = GRAYMATTER_THRESHOLD
    node_threshold.inputs.args = '-bin'

    node_graymatter = Node(function_fmri_graymatter, name='graymatter')

    node_realign_gm = Node(FLIRT(), name='realign_gm')
    node_realign_gm.inputs.apply_xfm = True
    node_realign_gm.inputs.uses_qform = True

    kernel_sizes = arange(
        PARAMETERS['at_elec']['kernel_start'],
        PARAMETERS['at_elec']['kernel_end'],
        PARAMETERS['at_elec']['kernel_step'],
        )
    node_atelec = Node(function_fmri_atelec, name='at_elec')
    node_atelec.inputs.distance = PARAMETERS['at_elec']['distance']
    node_atelec.inputs.kernel_sizes = list(kernel_sizes)
    node_atelec.inputs.graymatter = PARAMETERS['graymatter']

    w = Workflow('fmri')

    w.connect(input, 'T1w', node_bet, 'in_file')
    w.connect(input, 'bold', node_featdesign, 'func')
    w.connect(input, 'electrodes', node_atelec, 'electrodes')

    w.connect(node_bet, 'out_file', node_featdesign, 'anat')

    w.connect(node_featdesign, 'fsf_file', node_feat, 'fsf_file')
    w.connect(node_feat, 'feat_dir', node_compare, 'feat_path')

    if PARAMETERS['upsample']:
        w.connect(node_compare, 'out_file', node_upsample, 'in_file')
        w.connect(node_compare, 'out_file', node_upsample, 'reference')
        w.connect(node_upsample, 'out_file', node_atelec, 'in_file')
    else:
        w.connect(node_compare, 'out_file', node_atelec, 'in_file')

    if PARAMETERS['graymatter']:

        node_reconall = Node(ReconAll(), name='freesurfer')
        node_reconall.inputs.subjects_dir = str(FREESURFER_PATH)
        node_reconall.inputs.flags = ['-cw256', ]

        w.connect(input, 'T1w', node_reconall, 'T1_files')
        w.connect(input, 'subject', node_reconall, 'subject_id')

        if PARAMETERS['upsample']:
            w.connect(node_graymatter, 'out_file', node_realign_gm, 'in_file')
            w.connect(node_upsample, 'out_file', node_realign_gm, 'reference')
            w.connect(node_realign_gm, 'out_file', node_threshold, 'in_file')
        else:
            w.connect(node_graymatter, 'out_file', node_downsample, 'in_file')
            w.connect(node_compare, 'out_file', node_downsample, 'reference')
            w.connect(node_downsample, 'out_file', node_threshold, 'in_file')

        w.connect(node_threshold, 'out_file', node_atelec, 'graymatter')

        w.connect(node_reconall, 'ribbon', node_graymatter, 'ribbon')

    return w



from shutil import rmtree

from boavus.workflow.ieeg import workflow_ieeg
from nipype import config, logging

from .paths import (ANALYSIS_PATH,
                    BIDS_PATH,
                    task_ieeg,
                    elec,
                    parameters,
                    )


LOG_PATH = ANALYSIS_PATH / 'log'
config.update_config({
    'logging': {
        'log_directory': LOG_PATH,
        'log_to_file': True,
        },
    'execution': {
        'crashdump_dir': LOG_PATH,
        'keep_inputs': 'false',
        'remove_unnecessary_outputs': 'false',
        },
    })


def test_workflow_ieeg():

    w = workflow_ieeg(parameters['ieeg'])
    w.base_dir = str(ANALYSIS_PATH)

    node = w.get_node('input')
    node.inputs.ieeg = str(task_ieeg.get_filename(BIDS_PATH))
    node.inputs.electrodes = str(elec.get_filename(BIDS_PATH))

    w.write_graph(
        graph2use='flat',
        )

    rmtree(LOG_PATH, ignore_errors=True)
    LOG_PATH.mkdir()
    logging.update_logging(config)

    w.run()



"""Put in this module functions that might end up in bidso because they are
reused often.
"""

def find_labels_in_regions(electrodes, regions):

    if len(regions) > 0:
        select_regions = lambda x: x['region'] in regions
    else:
        select_regions = lambda x: True

    return electrodes.electrodes.get(filter_lambda=select_regions,
                                     map_lambda=lambda x: x['name'])


def read_channels(ecog_tsv, labels, COLUMN):
    """Read the columns from an ecog_tsv file, in the order of labels.

    Parameters
    ----------
    ecog_tsv : list of dict
        list of dictionary (such as '_compare.tsv')
    labels : list of str
        list of channels to read
    COLUMN : str
        which column of ecog_tsv to read

    Returns
    -------
    list of str
        list of channels with values
    list of float
        list of ecog_tsv values taken from COLUMN in the order specified by
        labels
    """
    good_labels = []
    vals = []
    for l in labels:
        try:
            vals.append([float(x[COLUMN]) for x in ecog_tsv if x['channel'] == l][0])
        except IndexError:  # no electrode in ecog_tsv
            continue
        else:
            good_labels.append(l)

    return good_labels, vals



import numpy as np
import nibabel as nib
from scipy.spatial.distance import cdist


def snap_elec_to_surf(e_init, surf_smooth):
    dura = nib.freesurfer.read_geometry(str(surf_smooth))[0]

    max_steps = 4000
    temperature_exponent = 1
    deformation_constant = 1.
    giveup_steps = 10000
    init_temp = 1e-3

    n = e_init.shape[0]
    alpha = np.zeros((n, n))
    init_dist = cdist(e_init, e_init)

    neighbors = []

    for i in range(n):
        neighbor_vec = init_dist[:, i]
        # take 5 highest neighbors
        i_neigh = np.min((5, len(neighbor_vec) - 1))
        h5, = np.where(np.logical_and(neighbor_vec < np.sort(neighbor_vec)[i_neigh],
                                      neighbor_vec != 0))

        neighbors.append(h5)

    neighbors = np.squeeze(neighbors)

    # get distances from each neighbor pairing
    neighbor_dists = []
    for i in range(n):
        neighbor_dists.append(init_dist[i, neighbors[i]])

    neighbor_dists = np.hstack(neighbor_dists)

    max = np.max(np.around(neighbor_dists))
    min = np.min(np.around(neighbor_dists))

    hist, _ = np.histogram(neighbor_dists, bins=int((max - min) / 2), range=(min, max))

    fundist = np.argmax(hist) * 2 + min + 1

    # apply fundist to alpha matrix
    alpha_tweak = 1.75

    for i in range(n):
        neighbor_vec = init_dist[:, i]
        neighbor_vec[i] = np.inf

        neighbors = np.where(neighbor_vec < fundist * alpha_tweak)

        if len(neighbors) > i_neigh:
            neighbors = np.where(neighbor_vec < np.sort(neighbor_vec)[i_neigh])

        if len(neighbors) == 0:
            closest = np.argmin(neighbors)
            neighbors = np.where(neighbor_vec < closest * alpha_tweak)

        alpha[i,neighbors]=1

        for j in range(i):
            if alpha[j,i]==1:
                alpha[i,j]=1
            if alpha[i,j]==1:
                alpha[j,i]=1

    # alpha is set, now do the annealing
    def energycost(e_new, e_old, alpha):
        n = len(alpha)

        dist_new = cdist(e_new, e_new)
        dist_old = cdist(e_old, e_old)

        H=0

        for i in range(n):
            H += deformation_constant*float(cdist( [e_new[i]], [e_old[i]] ))

            for j in range(i):
                H += alpha[i,j] * (dist_new[i,j] - dist_old[i,j])**2

        return H

    max_deformation = 3
    deformation_choice = 50

    # adjust annealing parameters
    # H determines maximal number of steps
    H = max_steps
    # Texp determines the steepness of temperateure gradient
    Texp = 1 - temperature_exponent / H
    # T0 sets the initial temperature and scales the energy term
    T0 = init_temp
    # Hbrk sets a break point for the annealing
    Hbrk = giveup_steps

    h = 0
    hcnt = 0
    lowcost = mincost = 1e6

    # start e-init as greedy snap to surface
    e_snapgreedy = dura[np.argmin(cdist(dura, e_init), axis=0)]

    e = np.array(e_snapgreedy).copy()
    emin = np.array(e_snapgreedy).copy()

    # TODO: there is a random element, and it seems to affect the results
    # we set a seed, otherwise the tests are not reproducible
    np.random.seed(0)

    # the annealing schedule continues until the maximum number of moves
    while h < H:
        h += 1
        hcnt += 1
        # terminate if no moves have been made for a long time
        if hcnt > Hbrk:
            break

        # current temperature
        T = T0 * (Texp**h)

        # select a random electrode
        e1 = np.random.randint(n)
        # transpose it with a *nearby* point on the surface

        # find distances from this point to all points on the surface
        dists = np.squeeze(cdist(dura, [e[e1]]))
        # take a distance within the minimum 5X

        # mindist = np.min(dists)
        mindist = np.sort(dists)[deformation_choice]
        candidate_verts, = np.where(dists < mindist * max_deformation)
        choice_vert = candidate_verts[np.random.randint(len(candidate_verts))]

        e_tmp = e.copy()
        e_tmp[e1] = dura[choice_vert]

        cost = energycost(e_tmp, e_init, alpha)

        if cost < lowcost or np.random.random() < np.exp(-(cost - lowcost) / T):
            e = e_tmp
            lowcost = cost

            if cost < mincost:
                emin = e
                mincost = cost
                hcnt = 0

            if mincost == 0:
                break

    return emin



"""Various functions, that I don't know where to place
"""
from nibabel import load as nload
from numpy import array, isnan
from nibabel import Nifti1Image
from pathlib import Path
from tempfile import mkstemp


def get_vox2ras_tkr(filename):
    """This should be identical
    mri_info --vox2ras-tkr filename
    """
    img = nload(str(filename))

    Nc, Nr, Ns = img.shape[:3]
    dC, dR, dS = img.header['pixdim'][1:4]
    vox2ras_tkr = array([
        [-dC, 0, 0, Nc / 2 * dC],
        [0, 0, dS, -Ns / 2 * dS],
        [0, -dR, 0, Nr / 2 * dR],
        [0, 0, 0, 1],
    ])

    return vox2ras_tkr


def ribbon2graymatter(ribbon_files, output_dir):
    graymatter_file = output_dir / 'graymatter.nii.gz'

    ribbon_rh_file = [x for x in ribbon_files if x.endswith('rh.ribbon.mgz')][0]
    ribbon_lh_file = [x for x in ribbon_files if x.endswith('lh.ribbon.mgz')][0]

    ribbon_rh = nload(str(ribbon_rh_file))
    ribbon_lh = nload(str(ribbon_lh_file))
    graymatter = ribbon_rh.get_data() + ribbon_lh.get_data()

    nifti = Nifti1Image(graymatter, ribbon_rh.affine)
    nifti.to_filename(str(graymatter_file))

    return graymatter_file


def mri_nan2zero(input_nii):
    """Remove NaN values and turn them into zeros (so that Freesurfer can handle
    them)

    Parameters
    ----------
    input_nii : Path
        path to nii.gz containing NaN

    Returns
    -------
    Path
        path to temporary nii.gz containing no NaN. You can remove the file
        with .unlink()
    """
    img = nload(str(input_nii))
    dat = img.get_data()
    dat[isnan(dat)] = 0
    img = Nifti1Image(dat, img.affine)

    tmp_nii = mkstemp(suffix='.nii.gz')[1]
    img.to_filename(tmp_nii)
    return Path(tmp_nii)



from numpy import (arange,
                   array,
                   dot,
                   Inf,
                   meshgrid,
                   pi,
                   )
from scipy.optimize import least_squares

from popeye.visual_stimulus import gaussian_2D


def modelfun(pp, dd, X, Y):

    g = gaussian_2D(X, Y, pp[0], pp[1], pp[2], pp[2], 0).flatten()
    g /= 2 * pi * pp[2] ** 2

    return pp[3] * dot(dd, g)


def fun_to_minimize(params, stimulus, X, Y, ydata):
    return modelfun(params, stimulus, X, Y) - ydata


def fit_analyzePRF(stimulus, data):
    res = stimulus.shape[0]

    X, Y = meshgrid(arange(res) + 1, arange(res) + 1)
    seed = array([
        (1 + res) / 2,
        (1 + res) / 2,
        res,
        1,
        ])

    bounds = array([
        [1 - res + 1, 1 - res + 1, 0, -Inf],
        [2 * res - 1, 2 * res - 1, Inf, Inf],
        ])

    stimulus = stimulus.reshape(res ** 2, -1, order='F').T

    result = least_squares(
        fun_to_minimize,
        x0=seed,
        bounds=bounds,
        method='trf',
        args=[stimulus, X, Y, data],
        )
    return result



from logging import getLogger
from numpy import array_equal, errstate, loadtxt, NaN

from nibabel import load as nload
from nibabel import save as nsave
from nibabel import Nifti1Image

from bidso.utils import replace_extension

lg = getLogger(__name__)


def compare_fmri(feat_path, measure, normalize_to_mean, output_dir):
    # measure='percent', normalize_to_mean=False):

    if measure == 'percent':
        fmri_stat = compute_percent(feat_path, normalize_to_mean)
    elif measure == 'zstat':
        fmri_stat = compute_zstat(feat_path)
    else:
        raise ValueError(f'Unknown measure: {measure}')

    task_path = output_dir / replace_extension(feat_path.name, '_compare.nii.gz')
    nsave(fmri_stat, str(task_path))

    return task_path


def compute_percent(feat_path, normalize_to_mean):
    """Calculate percent change for a task.

    Parameters
    ----------

    Returns
    -------
    instance of nibabel.Nifti1Image
        percent change as image
    """
    design = read_design(feat_path)

    pe_mri = nload(str(feat_path / 'stats' / 'pe1.nii.gz'))

    pe = pe_mri.get_data()
    pe[pe == 0] = NaN
    perc = pe * 100 * design.ptp()

    if normalize_to_mean:
        """I'm not sure if this is necessary, but for sure it increases the level
        of noise"""
        mean_mri = nload(str(feat_path / 'mean_func.nii.gz'))
        mean_func = mean_mri.get_data()
        array_equal(pe_mri.affine, mean_mri.affine)
        with errstate(invalid='ignore'):
            perc /= mean_func

    mask_mri = nload(str(feat_path / 'mask.nii.gz'))
    mask = mask_mri.get_data().astype(bool)
    perc[~mask] = NaN

    return Nifti1Image(perc, pe_mri.affine)


def compute_zstat(feat_path):
    return nload(str(feat_path / 'stats' / 'zstat1.nii.gz'))


def read_design(feat_path):
    """TODO: this could be a method of Feat"""
    return loadtxt(str(feat_path / 'design.mat'), skiprows=5)



from json import dump
from logging import getLogger
from multiprocessing import Pool
from numpy import array

from bidso import Electrodes
from bidso.find import find_in_bids, find_root
from bidso.utils import replace_underscore

from wonambi.attr import Freesurfer

from .elec.annealing import snap_elec_to_surf
from ..mri.surface import fill_surface

lg = getLogger(__name__)


def project_electrodes(elec, freesurfer, analysis_dir):

    elec.acquisition += 'projected'
    bids_root = find_root(elec.filename)
    tsv_electrodes = elec.get_filename(bids_root)

    elec_realigned = {}
    HEMI = {
        'L': 'lh',
        'R': 'rh',
    }

    groups = {x['group'] for x in elec.electrodes.tsv}
    for group in groups:
        print(group)
        anat_dir = analysis_dir / ('sub-' + elec.subject) / 'anat'
        anat_dir.mkdir(exist_ok=True, parents=True)
        labels = [x['name'] for x in elec.electrodes.tsv if x['group'] == group]
        hemi = [x['hemisphere'] for x in elec.electrodes.tsv if x['group'] == group][0]
        elec_type = [x['type'] for x in elec.electrodes.tsv if x['group'] == group][0]
        if not elec_type == 'surface':
            print(f'not realigning {group} because its type is {elec_type}')
            for _elec in elec.electrodes_tsv:
                if _elec['group'] == group:
                    elec_realigned[_elec['label']] = (float(_elec['x']), float(_elec['y']), float(_elec['z']))

        anat_file = anat_dir / (HEMI[hemi] + '.smooth')
        if not anat_file.exists():
            surf = getattr(freesurfer.read_brain('pial'), HEMI[hemi])
            fill_surface(surf.surf_file, anat_file)

        xyz = array([(float(x['x']), float(x['y']), float(x['z'])) for x in elec.electrodes.tsv if x['group'] == group])
        xyz -= freesurfer.surface_ras_shift
        xyz_realigned = snap_elec_to_surf(xyz, anat_file)

        for label, xyz_ in zip(labels, xyz_realigned):
            elec_realigned[label] = xyz_ + freesurfer.surface_ras_shift

    with tsv_electrodes.open('w') as f:
        f.write('name\tgroup\tx\ty\tz\tsize\ttype\tmaterial\themisphere\n')
        for _elec in elec.electrodes.tsv:
            label = _elec['name']
            xyz = "\t".join(f'{x:f}' for x in elec_realigned[label])
            one_chans = [x for x in elec.electrodes.tsv if x['name'] == label][0]
            group = one_chans['group']
            elec_type = one_chans['type']
            size = one_chans['size']
            material = one_chans['material']
            hemisphere = one_chans['hemisphere']
            f.write(f'{label}\t{group}\t{xyz}\t{size}\t{elec_type}\t{material}\t{hemisphere}\n')

    elec.coordframe.json['iEEGCoordinateProcessingDescription'] += '; Dijkstra et al.'  # TODO: better description + remove None
    new_json = replace_underscore(tsv_electrodes, 'coordsystem.json')
    with new_json.open('w') as f:
        dump(elec.coordframe.json, f, indent=2)

    return Electrodes(tsv_electrodes)



from pickle import load, dump
from logging import getLogger
from multiprocessing import Pool
import warnings
from numpy import (array,
                   diff,
                   empty,
                   mean,
                   percentile,
                   pi,
                   stack,
                   roll,
                   )

from scipy.signal import (iirdesign,
                          lfilter,
                          group_delay,
                          )
from scipy.stats.mstats import gmean

from bidso.find import find_in_bids
from bidso.utils import replace_extension

from wonambi.trans import math

lg = getLogger(__name__)


def main(analysis_dir, bands=[], method="5", noparallel=False):
    """
    extract timefrequency after bandpass

    Parameters
    ----------
    analysis_dir : path

    bands : str
        write down frequency bands as 70-90,90-110 (no spaces)
    method : str
        "1", "2", "3", "4", "5"
    noparallel : bool
        if it should run serially (i.e. not parallely, mostly for debugging)
    """
    # convert str "70-90,90-110" to [[70, 90], [90, 110]]
    bands = [[float(f) for f in b.split('-')] for b in bands.split(',')]

    args = []
    for ieeg_file in find_in_bids(analysis_dir, modality='ieegproc', extension='.pkl', generator=True):
        args.append((ieeg_file, bands, method))

    if noparallel:
        for arg in args:
            save_frequency(*arg)
    else:
        with Pool() as p:
            p.starmap(save_frequency, args)


def save_frequency(ieeg_file, bands, method):
    with ieeg_file.open('rb') as f:
        data = load(f)

    data = butterpass_eeglabdata(data, bands)
    data = extract_broadband(data, method)

    output_file = replace_extension(ieeg_file, 'broadband.pkl')
    with output_file.open('wb') as f:
        dump(data, f)


def butterpass_eeglabdata(data, bands=[]):
    data1 = data._copy()
    data1.axis['filter'] = empty(data.number_of('trial'), dtype='O')
    data1.data = empty(data.number_of('trial'), dtype='O')

    for i in range(data.number_of('trial')):
        data1.axis['filter'][i] = array([f'{bp[0]}-{bp[1]}' for bp in bands])
        trial = []
        for bp in bands:
            trial.append(butterpass_eeglabdata_core(data(trial=i), bp, data.s_freq))

        data1.data[i] = stack(trial, axis=2)

    return data1


def butterpass_eeglabdata_core(signal, band, srate, Rp=3, Rs=60, bw=.5):

    nyqLimit = srate / 2
    Fpass1 = band[0] / nyqLimit
    Fpass2 = band[1] / nyqLimit
    Fstop1 = Fpass1 * bw
    Fstop2 = Fpass2 / bw

    b, a = iirdesign([Fpass1, Fpass2], [Fstop1, Fstop2], Rp, Rs, ftype='butter', output='ba')
    y = lfilter(b, a, signal)

    # measure time shift of filter
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        f, gd = group_delay((b, a), int(srate), False)
    f = f * nyqLimit / pi
    shift_frames = int(mean(gd[(f > band[0]) & (f <= band[1])]))

    # correct for time shift of filter
    return roll(y, -shift_frames)


def whiten(x):
    return (x - x.mean()) / diff(percentile(x, [.25, .75]))


def extract_broadband(data, method):

    if method == '1':  # 'abs(hilbert)'
        data = math(data, operator=whiten)
        data = math(data, operator=mean, axis='filter')
        data = math(data, operator_name=['hilbert', 'abs'], axis='time')

    elif method == '2':  # 'abs(hilbert(bp))'
        data = math(data, operator=whiten)
        data = math(data, operator_name='mean', axis='filter')
        data = math(data, operator_name=['hilbert', 'abs', 'square'], axis='time')

    elif method == '3':  # 'abs(hilbert(sum(whiten(bp))))'
        data = math(data, operator=whiten)
        data = math(data, operator_name=['hilbert', 'abs'], axis='time')
        data = math(data, operator=gmean, axis='filter')

    elif method == '4':  # 'sum(abs(hilbert(whiten(bp))))'
        data = math(data, operator=whiten)
        data = math(data, operator_name=['hilbert', 'abs', 'square'], axis='time')
        data = math(data, operator=gmean, axis='filter')

    elif method == '5':  # 'sum(abs(hilbert(bp)))'
        data = math(data, operator_name=['hilbert', 'abs', 'square'], axis='time')
        data = math(data, operator=gmean, axis='filter')

    return data



from nipype import Workflow, Node
from nipype.interfaces.utility import IdentityInterface

from .fmri import workflow_fmri
from .ieeg import workflow_ieeg
from ..corr import function_corr


def workflow_corr_ieeg_fmri(PARAMETERS, FREESURFER_PATH):

    input = Node(IdentityInterface(fields=['subject', 'T1w', 'bold', 'ieeg', 'electrodes']), name='input')

    node_corr = Node(function_corr, name='corr_fmri_ecog')
    node_corr.inputs.pvalue = PARAMETERS['corr']['pvalue']

    w_fmri = workflow_fmri(PARAMETERS['fmri'], FREESURFER_PATH)
    w_ieeg = workflow_ieeg(PARAMETERS['ieeg'])

    w = Workflow('grvx')

    w.connect(input, 'ieeg', w_ieeg, 'input.ieeg')
    w.connect(input, 'electrodes', w_ieeg, 'input.electrodes')

    w.connect(input, 'subject', w_fmri, 'input.subject')
    w.connect(input, 'T1w', w_fmri, 'input.T1w')
    w.connect(input, 'bold', w_fmri, 'input.bold')
    w.connect(input, 'electrodes', w_fmri, 'input.electrodes')

    w.connect(w_ieeg, 'ecog_compare.tsv_compare', node_corr, 'ecog_file')
    w.connect(w_fmri, 'at_elec.fmri_vals', node_corr, 'fmri_file')

    return w



from pathlib import Path
from bidso import file_Core


TEST_PATH = Path(__file__).resolve().parent
DATA_PATH = TEST_PATH / 'data'

BIDS_PATH = DATA_PATH / 'bids'
FREESURFER_PATH = DATA_PATH / 'freesurfer'

ANALYSIS_PATH = DATA_PATH / 'analysis'
ANALYSIS_PATH.mkdir(parents=True, exist_ok=True)

subject = 'delft'
task_ieeg = file_Core(
    subject=subject,
    session='UMCUECOGday01',
    modality='ieeg',
    task='motorHandLeft',
    run='1',
    acquisition='clinical',
    extension='.eeg',
    )
task_fmri = file_Core(
    subject=subject,
    session='UMCU3Tdaym13',
    modality='bold',
    task='motorHandLeft',
    run='1',
    extension='.nii.gz',
    )
task_anat = file_Core(
    subject=subject,
    session='UMCU3Tdaym13',
    modality='T1w',
    acquisition='wholebrain',
    extension='.nii.gz',
    )
elec = file_Core(
    subject=subject,
    session='UMCUECOGday01',
    modality='electrodes',
    acquisition='clinical',
    extension='.tsv',
    )


parameters = {
    'corr': {
        'pvalue': 0.05,
        },
    'fmri': {
        "fmri_compare": {
            "measure": "zstat",
            "normalize_to_mean": True
            },
        "at_elec": {
            "distance": "gaussian",
            "kernel_start": 3,
            "kernel_end": 10,
            "kernel_step": 0.5
            },
        "upsample": False,
        "graymatter": True,
        },
    'ieeg': {
        'read': {
            'conditions': {
                'move': [49, ],
                'rest': [48, ],
                },
            'minimalduration': 1,
            },
        'preprocess': {
            'duration': 2,
            'reref': 'average',
            'offset': False,
            },
        'powerspectrum': {
            'method': 'spectrogram',
            'taper': 'hann',
            'duration': 2,
            },
        'ecog_compare': {
            'frequency': [65, 95],
            'baseline': True,
            'measure': 'zstat',
            'method': '3c',
            }
        }
    }



from nipype import Function


def wrapper_fmri_compare(feat_path, measure, normalize_to_mean):
    from pathlib import Path
    from boavus.fmri.compare import compare_fmri

    output = compare_fmri(
        Path(feat_path),
        measure,
        normalize_to_mean,
        Path('.').resolve())
    return str(output)


def wrapper_at_elec(in_file, electrodes, distance, kernel_sizes, graymatter=False):
    # measure_nii, electrodes, freesurfer_dir='', graymatter=False, distance='guassian', kernel_sizes):
    from pathlib import Path
    from boavus.fmri.at_electrodes import calc_fmri_at_elec

    output = calc_fmri_at_elec(
        Path(in_file),
        Path(electrodes),
        distance,
        kernel_sizes,
        graymatter,
        Path('.').resolve())
    return [str(x) for x in output]


def wrapper_fmri_graymatter(ribbon):
    from pathlib import Path
    from boavus.fmri.utils import ribbon2graymatter
    print(ribbon)

    output = ribbon2graymatter(
        ribbon,
        Path('.').resolve())
    return str(output)


function_fmri_compare = Function(
    input_names=[
        'feat_path',
        'measure',
        'normalize_to_mean',
    ],
    output_names=[
        'out_file',
    ],
    function=wrapper_fmri_compare,
    )


function_fmri_atelec = Function(
    input_names=[
        'in_file',
        'electrodes',
        'distance',
        'kernel_sizes',
        'graymatter',
    ],
    output_names=[
        'fmri_vals',
        'n_voxels',
    ],
    function=wrapper_at_elec,
    )


function_fmri_graymatter = Function(
    input_names=[
        'ribbon',
    ],
    output_names=[
        'out_file',
    ],
    function=wrapper_fmri_graymatter,
    )



from pathlib import Path
from nibabel import load as niload

from bidso import Task
from bidso.utils import replace_underscore, read_tsv, replace_extension

DESIGN_TEMPLATE = Path(__file__).resolve().parents[1] / 'data/design_template.fsf'


def prepare_design(func, anat, output_dir):
    """You should set remove_unnecessary_outputs to False, otherwise it removes
    the events.tsv file
    """

    task = Task(func)

    events_fsl = output_dir / task.events.filename.name
    _write_events(task.events.filename, events_fsl)

    # collect info
    img = niload(str(task.filename))
    n_vols = img.header.get_data_shape()[3]
    tr = img.header['pixdim'][4]  # Not sure it it's reliable

    with DESIGN_TEMPLATE.open('r') as f:
        design = f.read()

    feat_dir = output_dir / replace_extension(Path(task.filename).name, '.feat')

    design_values = {
        'XXX_OUTPUTDIR': str(feat_dir),
        'XXX_NPTS': str(n_vols),
        'XXX_TR': str(tr),
        'XXX_FEAT_FILE': str(task.filename),
        'XXX_HIGHRES_FILE': str(anat),
        'XXX_EV1': 'active',
        'XXX_TSVFILE': str(events_fsl),
        }

    for pattern, value in design_values.items():
        design = design.replace(pattern, value)

    subj_design = output_dir / replace_underscore(Path(task.filename).name, 'design.fsf')

    with subj_design.open('w') as f:
        f.write(design)

    return subj_design


def _write_events(events_input, events_output):
    """
    TODO
    ----
    EVENTS should be in PARAMETERS
    """
    EVENTS = (
        'move',
        'verbgen',
        )

    tsv = read_tsv(events_input)
    with events_output.open('w') as f:
        for event in tsv:
            if event['trial_type'] in EVENTS:
                f.write(f'{event["onset"]}\t{event["duration"]}\t1\n')



from nipype import Function

def wrapper_read_ieeg_block(ieeg, electrodes, conditions, minimalduration):
    from pathlib import Path
    from boavus.ieeg.read import read_ieeg_block

    outputs = read_ieeg_block(
        Path(ieeg),
        Path(electrodes),
        conditions,
        minimalduration,
        Path('.').resolve())
    return [str(x) for x in outputs]


def wrapper_preprocess(ieeg, reref, duration, offset):
    from pathlib import Path
    from boavus.ieeg.preprocessing import preprocess_ecog

    output = preprocess_ecog(
        Path(ieeg),
        reref,
        duration,
        offset,
        Path('.').resolve())
    return str(output)


def wrapper_powerspectrum(ieeg, method, taper, duration):
    from pathlib import Path
    from boavus.ieeg.psd import compute_powerspectrum

    output = compute_powerspectrum(
        Path(ieeg),
        method,
        taper,
        duration,
        Path('.').resolve())
    return str(output)


def wrapper_ieeg_compare(in_files, frequency, baseline=False, method='dh2012',
                         measure='dh2012_r2'):
    from pathlib import Path
    from boavus.ieeg.compare import compare_ieeg_freq

    output = compare_ieeg_freq(
        Path(in_files[0]),
        Path(in_files[1]),
        frequency,
        baseline,
        method,
        measure,
        Path('.').resolve())
    return str(output)


function_ieeg_read = Function(
    input_names=[
        'ieeg',
        'electrodes',
        'conditions',
        'minimalduration',
    ],
    output_names=[
        'ieeg',
    ],
    function=wrapper_read_ieeg_block,
    )


function_ieeg_preprocess = Function(
    input_names=[
        'ieeg',
        'reref',
        'duration',
        'offset',
    ],
    output_names=[
        'ieeg',
    ],
    function=wrapper_preprocess,
    )


function_ieeg_powerspectrum = Function(
    input_names=[
        'ieeg',
        'method',
        'taper',
        'duration',
    ],
    output_names=[
        'ieeg',
    ],
    function=wrapper_powerspectrum,
    )


function_ieeg_compare = Function(
    input_names=[
        'in_files',
        'frequency',
        'baseline',
        'method',
        'measure',
    ],
    output_names=[
        'tsv_compare',
    ],
    function=wrapper_ieeg_compare,
    )



from logging import getLogger
from pickle import load, dump
from numpy import empty, arange
from wonambi.trans import montage
from wonambi.trans.select import _create_subepochs

from bidso.utils import replace_extension


lg = getLogger(__name__)


def preprocess_ecog(ieeg_file, reref, duration, offset, output_dir):
    """

    Parameters
    ----------
    reref : str
        'average' or 'regression'
    duration : int
        length of the segments
    offset : bool
        remove one sample for whole duration

    TODO
    ----
    labels_in_roi = find_labels_in_regions(electrodes, regions)
    clean_roi_labels = [label for label in clean_labels if label in labels_in_roi]
    data = select(data, chan=clean_roi_labels)
    """
    with ieeg_file.open('rb') as f:
        data = load(f)

    data = montage(data, ref_to_avg=True, method=reref)
    data = make_segments(data, duration, offset)

    output_file = output_dir / replace_extension(ieeg_file.name, 'proc.pkl')
    with output_file.open('wb') as f:
        dump(data, f)

    return output_file


def make_segments(dat, duration=2, offset=True):

    dur_smp = int(dat.s_freq * duration)
    if offset:
        dur_smp -= 1

    trials = []
    for d in dat.data:
        v = _create_subepochs(d, dur_smp, dur_smp)
        for i in range(v.shape[1]):
            trials.append(v[:, i, :])

    out = dat._copy(axis=False)
    out.data = empty(len(trials), dtype='O')
    out.axis['chan'] = empty(len(trials), dtype='O')
    out.axis['time'] = empty(len(trials), dtype='O')

    for i, trial in enumerate(trials):
        out.data[i] = trial
        out.axis['time'][i] = arange(i * dur_smp, i * dur_smp + dur_smp) / dat.s_freq
        out.axis['chan'][i] = dat.axis['chan'][0]

    return out



from nipype import Function


def wrapper_prepare_design(func, anat):
    from pathlib import Path
    from boavus.fsl.feat import prepare_design

    output = prepare_design(
        Path(func),
        Path(anat),
        Path('.').resolve())
    return str(output)


function_prepare_design = Function(
    input_names=[
        'func',
        'anat',
    ],
    output_names=[
        'fsf_file',
    ],
    function=wrapper_prepare_design,
    )



from pathlib import Path
from shutil import copyfile
from subprocess import run, PIPE
from tempfile import mkstemp

from bidso.utils import bids_mkdir, replace_underscore, replace_extension

from ..utils import check_subprocess, ENVIRON


def run_reorient2std(nii):
    """This function simply reorients nifti, so that FSL can work with it
    more easily (reg works much better after running this function).
    """
    tmp_nii = mkstemp(suffix='.nii.gz')[1]
    p = run([
        'fslreorient2std',
        str(nii),
        tmp_nii,
        ], stdout=PIPE, stderr=PIPE, env=ENVIRON)
    check_subprocess(p)
    copyfile(tmp_nii, nii)
    Path(tmp_nii).unlink()


def run_flirt_resample(nii_in, nii_out, target_resolution):
    """Downsample and upsample mri
    """
    p = run([
        'flirt',
        '-in', str(nii_in),
        '-ref', str(nii_in),
        '-out', str(nii_out),
        '-applyisoxfm', str(target_resolution),
        ], stdout=PIPE, stderr=PIPE, env=ENVIRON)
    check_subprocess(p)


def run_flirt_feat(task_fmri, gm_file):
    gm_feat_file = replace_extension(gm_file, '_feat.nii.gz')  # TEST: same affine as bold compare
    p = run([
        'flirt',
        '-in', str(gm_file),
        '-ref', str(task_fmri.filename),
        '-usesqform',
        '-applyxfm',
        '-out', str(gm_feat_file),
        ], stdout=PIPE, stderr=PIPE, env=ENVIRON)
    check_subprocess(p)
    return gm_feat_file


def run_fslmaths_threshold(gm_file, threshold):
    gm_bin_file = replace_extension(gm_file, '_bin.nii.gz')
    p = run([
        'fslmaths',
        gm_file,
        '-thr', str(threshold),
        '-bin',
        gm_bin_file
        ], stdout=PIPE, stderr=PIPE)
    check_subprocess(p)
    return gm_bin_file



from functools import partial
from itertools import product
from logging import getLogger
from multiprocessing import Pool
import warnings

from numpy import (ndindex,
                   array,
                   sum,
                   power,
                   zeros,
                   NaN,
                   isfinite,
                   nansum,
                   isnan,
                   where,
                   )
from numpy.linalg import norm, inv
from scipy.stats import norm as normdistr
from nibabel.affines import apply_affine
from nibabel import load as nload

from bidso import Electrodes
from bidso.utils import replace_underscore


lg = getLogger(__name__)

COUNT_THRESHOLD = 0.5
# 1 sigma = 0.6065306597126334


def calc_fmri_at_elec(measure_nii, electrodes_file, distance, kernels, graymatter, output_dir):
    """
    Calculate the (weighted) average of fMRI values at electrode locations
    """
    electrodes = Electrodes(electrodes_file)

    img = nload(str(measure_nii))
    mri = img.get_data()
    mri[mri == 0] = NaN

    labels = electrodes.electrodes.get(map_lambda=lambda x: x['name'])
    chan_xyz = array(electrodes.get_xyz())

    nd = array(list(ndindex(mri.shape)))
    ndi = from_mrifile_to_chan(img, nd)

    if graymatter:
        gm_mri = nload(str(graymatter)).get_data().astype(bool)
        mri[~gm_mri] = NaN

    lg.debug(f'Computing fMRI values for {measure_nii.name} at {len(labels)} electrodes and {len(kernels)} "{distance}" kernels')
    fmri_vals, n_voxels = compute_kernels(kernels, chan_xyz, mri, ndi, distance)

    fmri_vals_tsv = output_dir / replace_underscore(measure_nii.name, 'compare.tsv')
    n_voxels_tsv = output_dir / replace_underscore(measure_nii.name, 'nvoxels.tsv')

    with fmri_vals_tsv.open('w') as f:
        f.write('channel\t' + '\t'.join(str(one_k) for one_k in kernels) + '\n')
        for one_label, val_at_elec in zip(labels, fmri_vals):
            f.write(one_label + '\t' + '\t'.join(str(one_val) for one_val in val_at_elec) + '\n')

    with n_voxels_tsv.open('w') as f:
        f.write('channel\t' + '\t'.join(str(one_k) for one_k in kernels) + '\n')
        for one_label, val_at_elec in zip(labels, n_voxels):
            f.write(one_label + '\t' + '\t'.join(str(one_val) for one_val in val_at_elec) + '\n')

    return fmri_vals_tsv, n_voxels_tsv


def from_chan_to_mrifile(img, xyz):
    return apply_affine(inv(img.affine), xyz).astype(int)


def from_mrifile_to_chan(img, xyz):
    return apply_affine(img.affine, xyz)


def compute_kernels(kernels, chan_xyz, mri, ndi, distance, n_cpu=None):
    partial_compute_chan = partial(compute_chan, ndi=ndi, mri=mri, distance=distance)

    args = product(chan_xyz, kernels)
    if n_cpu is None:
        output = [partial_compute_chan(*arg) for arg in args]
    else:
        lg.debug(f'Number of CPU: {n_cpu}')
        with Pool(n_cpu) as p:
            output = p.starmap(partial_compute_chan, args)

    fmri_vals = [i[0] for i in output]
    n_voxels = [i[1] for i in output]

    fmri_vals = array(fmri_vals).reshape(-1, len(kernels))
    n_voxels = array(n_voxels).reshape(-1, len(kernels))

    return fmri_vals, n_voxels


def compute_chan(pos, KERNEL, ndi, mri, distance):
    dist_chan = norm(ndi - pos, axis=1)

    if distance == 'gaussian':
        m = normdistr.pdf(dist_chan, scale=KERNEL)
        m /= normdistr.pdf(0, scale=KERNEL)  # normalize so that peak is at 1, so that it's easier to count voxels

    elif distance == 'sphere':
        m = zeros(dist_chan.shape)
        m[dist_chan <= KERNEL] = 1

    elif distance == 'inverse':
        m = power(dist_chan, -1 * KERNEL)

    m = m.reshape(mri.shape)
    m[isnan(mri)] = NaN
    n_vox = _count_voxels(m)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        m /= sum(m[isfinite(m)])  # normalize so that the sum of the finite numbers is 1

    mq = m * mri
    return nansum(mq), n_vox


def _count_voxels(m):
    """Count voxels inside a 3D weight matrix.

    Parameters
    ----------
    m : 3d ndarray
        matrix as same size as mri, where the peak is 1 and each voxel has
        the weight based on the distance parameters

    Returns
    -------
    int
        number of voxels which have a weight higher than COUNT_THRESHOLD

    Notes
    -----
    Suppress RunTimeWarning because of > used against NaN.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        n_vox = len(where(m >= COUNT_THRESHOLD)[0])

    return n_vox



"""Complex but elegant approach to passing arguments and parameters.

Make sure that the functions are organized in module/function.py.
"""
from inspect import signature, getdoc
from collections import defaultdict
from argparse import ArgumentParser, RawTextHelpFormatter
from importlib import import_module
from logging import getLogger, StreamHandler, Formatter, INFO, DEBUG
from pathlib import Path
from warnings import filterwarnings

filterwarnings('ignore', category=FutureWarning)
lg = getLogger('boavus')

FOLDERS_DOC = {
    'bids_dir': 'The directory with the raw data organized in the BIDS format',
    'freesurfer_dir': 'The directory with Freesurfer',
    'analysis_dir': 'The directory with preprocessed / analyzed data for each subject',
    'output_dir': 'The directory with custom output',
    }


def __main__():
    import boavus
    all_mod = defaultdict(dict)
    for one in sorted(Path(boavus.__path__[0]).rglob('*.py')):
        mod = import_module(str(one.relative_to(Path(__file__).parents[1])).replace('/', '.')[:-3])
        if hasattr(mod, 'main'):
            mod_name = one.stem
            grp_name = one.parent.stem
            all_mod[grp_name][mod_name] = mod.main

    parser = ArgumentParser(description='Tools to analyze data structured as BIDS in Python', formatter_class=RawTextHelpFormatter)
    list_modules = parser.add_subparsers(title='Modules', help='Modules containing the functions')

    for m_k, m_v in all_mod.items():
        module = list_modules.add_parser(m_k)
        module.set_defaults(module=m_k)
        list_functions = module.add_subparsers(title=f'Functions in {m_k} module')

        for f_k, f_v in m_v.items():
            function = list_functions.add_parser(f_k,
                                                 formatter_class=RawTextHelpFormatter)
            function.set_defaults(function=f_k)
            add_to_parser(function, f_v)
            function.add_argument('-l', '--log', default='info',
                                  help='Logging level: info (default), debug')

    return parser


def add_to_parser(function, main_f):
    doc = getdoc(main_f)
    args = {}
    maindoc, docargs = doc.split('\nParameters\n----------\n')
    docargs = docargs.split('\n')
    for i in range(0, len(docargs), 2):
        a_name, a_type = docargs[i].split(':')
        if i < len(docargs) - 1:
            comment = docargs[i + 1].strip()
        else:
            comment = ''
        args[a_name.strip()] = (a_type.strip(), comment)

    sign = signature(main_f)

    folders_arg = function.add_argument_group('folders arguments')
    optionals_arg = function.add_argument_group('optional arguments')

    for name, param in sign.parameters.items():
        if args[name][0] == 'path':
            one_arg = folders_arg
            help_str = FOLDERS_DOC[name]
            metavar = name.upper()
            action = None
            const = None

        elif args[name][0] == 'bool':
            assert not param.default, f'bool value {name} should be False, otherwise it\'s hard to interpret'
            one_arg = optionals_arg
            help_str = args[name][1]
            metavar = None
            action = 'store_const'
            const = not param.default

        else:
            one_arg = optionals_arg
            help_str = args[name][1]
            metavar = f'"{param.default}"'
            action = None
            const = None

        if param.default == param.empty:
            required = True
        else:
            required = False
        if not required and args[name][0] == 'path':
            help_str = '(optional) ' + help_str

        one_arg.add_argument(
            '--' + param.name,
            required=required,
            help=help_str,
            metavar=metavar,
            action=action,
            const=const,
            default=param.default,
            )


def boavus(arguments=None):

    parser = __main__()

    # treat arguments as dict so we can remove keys as we go
    args = vars(parser.parse_args(arguments))

    # log can be info or debug
    DATE_FORMAT = '%H:%M:%S'
    log_level = args.pop('log')
    if log_level[:1].lower() == 'i':
        lg.setLevel(INFO)
        FORMAT = '{asctime:<10}{message}'

    elif log_level[:1].lower() == 'd':
        lg.setLevel(DEBUG)
        FORMAT = '{asctime:<10}{levelname:<10}{filename:<40}(l. {lineno: 6d})/ {funcName:<40}: {message}'

    formatter = Formatter(fmt=FORMAT, datefmt=DATE_FORMAT, style='{')
    handler = StreamHandler()
    handler.setFormatter(formatter)

    lg.handlers = []
    lg.addHandler(handler)

    """
    the user passes two arguments: a module and a function. A module is the
    name of the folder and the function is the name of the .py file.
    Each .py file contains one and ony one main() function, which is what is
    called in general.
    """
    module = import_module('boavus.' + args.pop('module') + '.' + args.pop('function'))

    """
    Use the leftover arguments only
    """
    # this only works if all the arguments are paths
    for k, v in args.items():
        if k.endswith('_dir'):
            args[k] = _path(v)

    """
    Call the actual main function of the specified .py file
    """
    lg.debug(f'Calling main() from {module.__file__} with: ' + ', '.join(f'{k}={v}' for k, v in args.items()))
    module.main(**args)


def _path(dirname):
    """Always use absolute paths, easier to control when working with FSL / Freesurfer"""
    if dirname is None:
        return None
    else:
        return Path(dirname).resolve()



from logging import getLogger
from shutil import rmtree
from numpy import array, median

from bidso import Electrodes
from bidso.find import find_in_bids
from bidso.utils import replace_underscore, read_tsv

from wonambi.attr import Channels, Freesurfer
from wonambi.viz import Viz3

from ..bidso import read_channels

lg = getLogger(__name__)

ELECSURF_DIR = 'electrodes_on_surface'


def main(bids_dir, analysis_dir, freesurfer_dir, output_dir,
         acquisition='clinical', measure_modality="", measure_column=""):
    """
    plot electrodes onto the brain surface,

    Parameters
    ----------
    bids_dir : path

    analysis_dir : path

    freesurfer_dir : path

    output_dir : path

    acquisition : str
        acquisition type of the electrode files
    measure_modality : str
        modality
    measure_column : str
        column
    """
    img_dir = output_dir / ELECSURF_DIR
    rmtree(img_dir, ignore_errors=True)
    img_dir.mkdir(exist_ok=True, parents=True)

    for electrode_path in find_in_bids(bids_dir, generator=True, acquisition=acquisition, modality='electrodes', extension='.tsv'):
        lg.debug(f'Reading electrodes from {electrode_path}')
        elec = Electrodes(electrode_path)
        fs = Freesurfer(freesurfer_dir / ('sub-' + elec.subject))

        labels = None
        if measure_modality != "":
            try:
                ecog_file = find_in_bids(
                    analysis_dir,
                    subject=elec.subject,
                    modality=measure_modality,
                    extension='.tsv')
            except FileNotFoundError as err:
                lg.warning(err)
                continue

            lg.debug(f'Reading {measure_column} from {ecog_file}')
            ecog_tsv = read_tsv(ecog_file)

            labels = [x['name'] for x in elec.electrodes.tsv]
            labels, vals = read_channels(ecog_tsv, labels, measure_column)

        else:
            vals = None

        v = plot_electrodes(elec, fs, labels, vals)

        png_file = img_dir / replace_underscore(elec.get_filename(), 'surfaceplot.png')
        lg.debug(f'Saving electrode plot on {png_file}')
        v.save(png_file)
        v.close()


def plot_electrodes(elec, freesurfer, labels=None, values=None):
    if labels is None:
        labels = [x['name'] for x in elec.electrodes.tsv]
    xyz = array(elec.get_xyz(labels))

    # convert from RAS to tkRAS
    xyz -= freesurfer.surface_ras_shift

    chan = Channels(labels, xyz)
    if median(chan.return_xyz()[:, 0]) > 0:
        surf = freesurfer.read_brain().rh
    else:
        surf = freesurfer.read_brain().lh
    v = Viz3()
    v.add_chan(chan, values=values)
    v.add_surf(surf, alpha=0.5)

    return v



"""Interface to popeye
"""
from ctypes import c_int16
from numpy import (array,
                   pi,
                   tan,
                   )
from popeye.utilities import grid_slice
from popeye.visual_stimulus import VisualStimulus
from popeye.og import GaussianFit, GaussianModel


SCREEN_WIDTH = 25
N_PIXELS = 100
VIEWING_DISTANCE = SCREEN_WIDTH / (tan(pi / 360 * N_PIXELS) * 2)


def fit_popeye(bars, dat):

    stimulus = generate_stimulus(bars)
    model = generate_model(stimulus)

    # set search grid
    x_grid = grid_slice(-10, 10, 5)
    y_grid = grid_slice(-10, 10, 5)
    s_grid = grid_slice(0.25, 5.25, 5)

    # set search bounds
    x_bound = (-12.0, 12.0)
    y_bound = (-12.0, 12.0)
    s_bound = (0.001, 12.0)
    b_bound = (1e-8, None)
    m_bound = (None, None)

    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid,)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)

    # fit the response
    fit = GaussianFit(model, dat, grids, bounds)

    return fit


def generate_stimulus(bars):
    tr_length = 1.0
    scale_factor = 1.0
    dtype = c_int16
    stimulus = VisualStimulus(bars, VIEWING_DISTANCE, SCREEN_WIDTH, scale_factor, tr_length, dtype)

    return stimulus


def nohrf(*args):
    return array([1, ])


def generate_model(stimulus):
    model = GaussianModel(stimulus, nohrf)
    model.hrf_delay = 0
    model.mask_size = 6
    return model



from os import environ, pathsep
from subprocess import CompletedProcess

def _remove_python3_from_PATH(path):
    return pathsep.join(x for x in path.split(pathsep) if 'miniconda' not in x and 'venv/bin' not in x and 'python3' not in x)


ENVIRON = {
    'PATH': _remove_python3_from_PATH(environ.get('PATH', '')),
    }
ENVIRON = {**environ, **ENVIRON}


def check_subprocess(p):
    if p.returncode:

        if isinstance(p, CompletedProcess):
            stdout = p.stdout
            stderr = p.stderr
        else:
            stdout, stderr = p.communicate()

        raise RuntimeError(f'Command \'{" ".join(p.args)}\' failed:\n---stdout---\n{stdout.decode()}------------\n---stderr---\n{stderr.decode()}------------')



from json import dump
from multiprocessing import Pool
from numpy import array

from bidso import Electrodes
from bidso.find import find_in_bids, find_root
from bidso.utils import replace_underscore

from wonambi.attr import Freesurfer


def main(bids_dir, freesurfer_dir, acquisition='clinical', noparallel=False):
    """
    assign electrodes to brain regions',

    Parameters
    ----------
    bids_dir : path

    freesurfer_dir : path

    acquisition : str
        acquisition type of the electrode files
    noparallel : bool
        if it should run serially (i.e. not parallely, mostly for debugging)
    """
    args = []
    for electrode_path in find_in_bids(bids_dir, generator=True, acquisition=acquisition, modality='electrodes', extension='.tsv'):
        elec = Electrodes(electrode_path)
        fs = Freesurfer(freesurfer_dir / ('sub-' + elec.subject))
        args.append((elec, fs))

    if noparallel:
        for arg in args:
            assign_regions(*arg)
    else:
        with Pool(processes=4) as p:
            p.starmap(assign_regions, args)


def assign_regions(elec, freesurfer):
    bids_dir = find_root(elec.filename)
    elec.acquisition += 'regions'
    tsv_electrodes = elec.get_filename(bids_dir)

    with tsv_electrodes.open('w') as f:
        f.write('name\tx\ty\tz\ttype\tsize\tmaterial\tregion\n')  # TODO: region is not in BEP010
        for _tsv in elec.electrodes.tsv:
            xyz = array([float(_tsv['x']), float(_tsv['y']), float(_tsv['z'])])
            region = freesurfer.find_brain_region(xyz, exclude_regions=('White', 'WM', 'Unknown'))[0]
            f.write(f'{_tsv["name"]}\t{_tsv["x"]}\t{_tsv["y"]}\t{_tsv["z"]}\t{_tsv["type"]}\t{_tsv["size"]}\t{_tsv["material"]}\t{region}\n')

    elec.coordframe.json['iEEGCoordinateProcessingDescription'] += '; Assign brain regions'  # TODO: better description + remove None
    new_json = replace_underscore(tsv_electrodes, 'coordsystem.json')
    with new_json.open('w') as f:
        dump(elec.coordframe.json, f, indent=2)

    return Electrodes(tsv_electrodes)



from logging import getLogger
from shutil import rmtree
from subprocess import Popen, PIPE

from nibabel import load as nload
from wonambi.viz import Viz3
from wonambi.attr import Freesurfer

from bidso import file_Core
from bidso.find import find_in_bids
from bidso.utils import replace_underscore

from ..fmri.utils import mri_nan2zero
from ..utils import ENVIRON, check_subprocess


lg = getLogger(__name__)

SURF_DIR = 'feat_surf'


def main(analysis_dir, freesurfer_dir, output_dir, modality='compare',
         surface='white', surf_fwhm=0):
    """
    map feat values on freesurfer surface',

    Parameters
    ----------
    analysis_dir : path

    freesurfer_dir : path

    output_dir : path

    modality : str
        "compare"
    surface : str
        "white", "pial"
    surf_fwhm : float
        FWHM
    """
    p_all = []
    surfs = []
    for in_vol_file in find_in_bids(analysis_dir, generator=True, extension='.nii.gz', modality=modality):
        in_vol = file_Core(in_vol_file)
        feat_path = find_in_bids(analysis_dir, subject=in_vol.subject, extension='.feat')
        for hemi in ('lh', 'rh'):
            p, out_surf = vol2surf(in_vol, feat_path, freesurfer_dir, hemi, surface, surf_fwhm)
            p_all.append(p)
            surfs.append(out_surf)

    # wait for all processes to run
    [p.wait() for p in p_all]
    [check_subprocess(p) for p in p_all]
    [info['mri_nonan'].unlink() for info in surfs]

    img_dir = output_dir / SURF_DIR
    rmtree(img_dir, ignore_errors=True)
    img_dir.mkdir(exist_ok=True, parents=True)

    for one_surf in surfs:
        plot_surf(img_dir, freesurfer_dir, one_surf, surface)


def vol2surf(in_vol, feat_path, freesurfer_dir, hemi, surface, surf_fwhm):
    out_surf = replace_underscore(in_vol.filename, in_vol.modality + 'surf' + hemi + '.mgh')
    mri_nonan = mri_nan2zero(in_vol.filename)

    cmd = [
        'mri_vol2surf',
        '--src',
        str(mri_nonan),
        '--srcreg',
        str(feat_path / 'reg/freesurfer/anat2exf.register.dat'),
        '--trgsubject',
        'sub-' + in_vol.subject,  # freesurfer subject
        '--hemi',
        hemi,
        '--out',
        str(out_surf),
        '--surf',
        surface,
        '--surf-fwhm',
        str(surf_fwhm),
        ]

    p = Popen(cmd, env={**ENVIRON, 'SUBJECTS_DIR': str(freesurfer_dir)},
              stdout=PIPE, stderr=PIPE)

    info = {
        "surf": out_surf,
        "hemi": hemi,
        "subject": 'sub-' + in_vol.subject,
        "mri_nonan": mri_nonan,
        }

    return p, info


def plot_surf(img_dir, freesurfer_dir, info, surface):

    fs = Freesurfer(freesurfer_dir / info['subject'])
    surf = getattr(fs.read_brain(surface), info['hemi'])

    surf_img = nload(str(info['surf']))
    surf_val = surf_img.get_data()[:, 0, 0].astype('float64')

    v = Viz3()
    v.add_surf(surf, values=surf_val, limits_c=(-6, 6), colorbar=True)
    v.save(img_dir / (info['surf'].stem + '.png'))
    v.close()



from logging import getLogger

from numpy import argmax, polyfit, isnan, NaN, isin
from numpy.testing import assert_array_equal
from scipy.stats import linregress

from bidso.utils import read_tsv, replace_underscore


lg = getLogger(__name__)


def compute_corr_ecog_fmri(fmri_file, ecog_file, output_dir, PVALUE):
    output_dir.mkdir(exist_ok=True, parents=True)

    fmri_tsv = read_tsv(fmri_file)
    ecog_tsv = read_tsv(ecog_file)
    fmri_tsv = select_channels(fmri_tsv, ecog_tsv)
    kernel_sizes = fmri_tsv.dtype.names[1:]

    results_tsv = output_dir / replace_underscore(ecog_file.stem, 'bold_r2.tsv')
    with results_tsv.open('w') as f:
        f.write('Kernel\tRsquared\tSlope\tIntercept\n')

        for kernel in kernel_sizes:
            try:
                r2, slope, intercept = compute_rsquared(
                    ecog_tsv['measure'],
                    fmri_tsv[kernel],
                    ecog_tsv['pvalue'],
                    PVALUE)

            except Exception:
                r2 = slope = intercept = NaN

            f.write(f'{float(kernel):.2f}\t{r2}\t{slope}\t{intercept}\n')

    return results_tsv


def select_channels(fmri_vals, ecog_vals):
    """make sure we're using the same channels and in the same order
    """
    fmri_idx = isin(fmri_vals['channel'], (ecog_vals['channel']))
    fmri_vals = fmri_vals[fmri_idx]
    assert_array_equal(fmri_vals['channel'], ecog_vals['channel'])

    return fmri_vals


def compute_rsquared(x, y, p_val, PVALUE):
    mask = ~isnan(x) & ~isnan(y) & (p_val <= PVALUE)

    lr = linregress(x[mask], y[mask])
    return lr.rvalue ** 2, lr.slope, lr.intercept


def read_shape(one_tsv):
    results = read_tsv(one_tsv)
    k = [float(x['Kernel']) for x in results]
    rsquared = [float(x['Rsquared']) for x in results]

    return polyfit(k, rsquared, 2)[0], k[argmax(rsquared)], max(rsquared)



from difflib import SequenceMatcher
from pickle import load
from numpy import ones, hstack, sign, array, NaN
from numpy import concatenate as np_concatenate
from scipy.stats import ttest_ind, pearsonr

from wonambi.trans import math, concatenate, select
from wonambi.datatype import Data

from bidso import file_Core


def compare_ieeg_freq(file_A, file_B, frequency, baseline, merge_method, measure,
                      output_dir):
    """
    Parameters
    ----------
    baseline : bool
        if you want to substract baseline
    merge_method : str
        "dh2012"
    measure : str
        "dh2012_r2"
    """
    ieeg_A = file_Core(file_A)
    ieeg_B = file_Core(file_B)

    with file_A.open('rb') as f:
        dat_A = load(f)
    with file_B.open('rb') as f:
        dat_B = load(f)

    if baseline:
        dat_A, dat_B = correct_baseline(dat_A, dat_B, frequency)

    hfa_A = merge(dat_A, merge_method, frequency)
    hfa_B = merge(dat_B, merge_method, frequency)

    if measure == 'diff':
        ecog_stats = compute_diff(hfa_A, hfa_B)
    elif measure == 'percent':
        ecog_stats = compute_percent(hfa_A, hfa_B)
    elif measure in ('zstat', 'dh2012_t'):  # identical
        ecog_stats = compute_zstat(hfa_A, hfa_B)
        if measure == 'dh2012_t':
            ecog_stats.data[0] *= -1  # opposite sign in dh2012's script

    elif measure == 'dh2012_r2':
        ecog_stats = calc_dh2012_values(hfa_A, hfa_B, measure)

    # need to check pvalues
    if True:
        pvalues = calc_dh2012_values(hfa_A, hfa_B, 'dh2012_pv')
    else:
        pvalues = [NaN, ] * ecog_stats.number_of('chan')[0]

    output = file_Core(
        subject=ieeg_A.subject,
        session=ieeg_A.session,
        run=ieeg_A.run,
        acquisition=ieeg_A.acquisition,
        modality='compare',
        extension='.tsv',
        task=find_longest_match(ieeg_A.task, ieeg_B.task),
        )
    compare_file = output_dir / output.get_filename()
    with compare_file.open('w') as f:
        f.write('channel\tmeasure\tpvalue\n')
        for i, chan in enumerate(ecog_stats.chan[0]):
            f.write(f'{chan}\t{ecog_stats(trial=0, chan=chan)}\t{pvalues(trial=0, chan=chan)}\n')

    return compare_file


def merge(freq, method, frequency):

    freq = select(freq, freq=frequency)

    if method == '1a':
        freq = concatenate(freq, axis='time')
        freq = math(freq, operator_name='mean', axis='time')
        freq = math(freq, operator_name='mean', axis='freq')
        # only one value
        out = Data(freq.data[0][:, None], freq.s_freq, chan=freq.chan[0], time=(0, ))

    elif method == '1b':
        freq = concatenate(freq, axis='time')
        freq = math(freq, operator_name='dB')
        freq = math(freq, operator_name='mean', axis='freq')
        freq = math(freq, operator_name='mean', axis='time')
        # only one value
        out = Data(freq.data[0][:, None], freq.s_freq, chan=freq.chan[0], time=(0, ))

    elif method == '1c':
        freq = concatenate(freq, axis='time')
        freq = math(freq, operator_name='mean', axis='freq')
        freq = math(freq, operator_name='dB')
        freq = math(freq, operator_name='mean', axis='time')
        # only one value
        out = Data(freq.data[0][:, None], freq.s_freq, chan=freq.chan[0], time=(0, ))

    elif method == '1d':
        freq = concatenate(freq, axis='time')
        freq = math(freq, operator_name='mean', axis='freq')
        freq = math(freq, operator_name='mean', axis='time')
        freq = math(freq, operator_name='dB')
        # only one value
        out = Data(freq.data[0][:, None], freq.s_freq, chan=freq.chan[0], time=(0, ))

    elif method == '2a':
        freq = math(freq, operator_name='mean', axis='time')
        freq = math(freq, operator_name='mean', axis='freq')
        # one value per trial
        out = concatenate(freq, axis='trial')

    elif method == '2b':
        freq = math(freq, operator_name='dB')
        freq = math(freq, operator_name='mean', axis='time')
        freq = math(freq, operator_name='mean', axis='freq')
        # one value per trial
        out = concatenate(freq, axis='trial')

    elif method == '2c':
        freq = math(freq, operator_name='mean', axis='time')
        freq = math(freq, operator_name='dB')
        freq = math(freq, operator_name='mean', axis='freq')
        # one value per trial
        out = concatenate(freq, axis='trial')

    elif method == '2d':
        freq = math(freq, operator_name='mean', axis='time')
        freq = math(freq, operator_name='mean', axis='freq')
        freq = math(freq, operator_name='dB')
        # one value per trial
        out = concatenate(freq, axis='trial')

    elif method == '3a':
        freq = concatenate(freq, axis='time')
        # values per time point
        out = math(freq, operator_name='mean', axis='freq')

    elif method == '3b':
        freq = concatenate(freq, axis='time')
        freq = math(freq, operator_name='dB')
        # values per time point
        out = math(freq, operator_name='mean', axis='freq')

    elif method == '3c':
        freq = concatenate(freq, axis='time')
        freq = math(freq, operator_name='mean', axis='freq')
        # values per time point
        out = math(freq, operator_name='dB')

    elif method == 'dh2012':
        # identical to 3b, but use log instead of dB
        freq = concatenate(freq, axis='time')
        freq = math(freq, operator_name='log')
        # values per time point
        out = math(freq, operator_name='mean', axis='freq')

    return out


def compute_diff(hfa_A, hfa_B):
    hfa_A.data[0] -= hfa_B.data[0]
    return Data(hfa_A.data[0][:, 0], hfa_A.s_freq, chan=hfa_A.chan[0])


def compute_percent(hfa_A, hfa_B):
    x_A = math(hfa_A, operator_name='mean', axis=hfa_A.list_of_axes[1])
    x_B = math(hfa_B, operator_name='mean', axis=hfa_A.list_of_axes[1])

    perc = (x_A(trial=0) - x_B(trial=0)) / x_B(trial=0) * 100
    data_perc = Data(perc, hfa_A.s_freq, chan=hfa_A.chan[0])

    return data_perc


def compute_zstat(hfa_A, hfa_B):
    """
    TODO
    ----
    You can compute zstat by taking diff and then divide by standard deviation
    """
    zstat = ttest_ind(hfa_A.data[0], hfa_B.data[0], axis=1, equal_var=False).statistic

    return Data(zstat, hfa_A.s_freq, chan=hfa_A.chan[0])


def calc_dh2012_values(hfa_A, hfa_B, measure):
    """This is the exact translation of dh2012's Matlab code
    """
    ecog = hstack((hfa_A.data[0], hfa_B.data[0]))
    stim = hstack((ones(hfa_A.data[0].shape[1]), ones(hfa_B.data[0].shape[1]) * 0))

    val = []
    for ecog_chan in ecog:
        [r, p] = pearsonr(ecog_chan, stim)

        if measure == 'dh2012_r2':
            val.append(r ** 2 * sign(r))

        elif measure == 'dh2012_pv':
            val.append(p)

    return Data(array(val), hfa_A.s_freq, chan=hfa_A.chan[0])


def correct_baseline(freq_A, freq_B, frequency):
    move = select(freq_A, freq=frequency)
    rest = select(freq_B, freq=frequency)

    merged = merge_datasets(move, rest)
    merged = concatenate(merged, 'time')
    baseline = math(merged, operator_name='mean', axis='time')

    move.data[0] /= baseline.data[0][:, None, :]
    rest.data[0] /= baseline.data[0][:, None, :]
    return move, rest


def merge_datasets(dat1, dat2):
    both = dat1._copy(axis=False)
    both.data = np_concatenate((dat1.data, dat2.data))
    both.axis['time'] = np_concatenate((dat1.time, dat2.time))
    both.axis['chan'] = np_concatenate((dat1.chan, dat2.chan))
    both.axis['freq'] = np_concatenate((dat1.freq, dat2.freq))
    return both


def find_longest_match(taskA, taskB):
    s = SequenceMatcher(a=taskA, b=taskB).find_longest_match(0, len(taskA), 0, len(taskB))
    return taskA[s.a:(s.a + s.size)]



from pickle import load
from logging import getLogger
from multiprocessing import Pool

from bidso import Task
from bidso.find import find_in_bids
from bidso.utils import replace_extension
from numpy import where, array, zeros
from wonambi.trans import select, math, concatenate

from nibabel import load as nload
from nibabel.affines import apply_affine
from nibabel import Nifti1Image

from sanajeh.data import data_aparc
from sanajeh.fmri import select_region
from .core.least_squares import fit_analyzePRF
from .core.popeye import fit_popeye
# from ..ieeg.read import read_prf_stimuli

lg = getLogger(__name__)


def main(analysis_dir, method="analyzePRF", task='bairprf', input='ieegprocpsd', noparallel=False):
    """
    compute psd for two conditions

    Parameters
    ----------
    analysis_dir : path

    method : str
        "popeye" or "analyzePRF"
    task : str
        task to analyze
    input : str
        name of the modality of the preceding step
    noparallel : bool
        if it should run serially (i.e. not parallely, mostly for debugging)
    """
    args = []
    for prf_file in find_in_bids(analysis_dir, task=task, modality=input, extension='.pkl', generator=True):
        if input.startswith('ieeg'):
            funct = estimate_ieeg_prf

        elif input.startswith('bold'):
            funct = estimate_bold_prf

        else:
            raise ValueError(f'Unknown modality {input}')

        args.append((funct, prf_file, method))

    if noparallel:
        for arg in args:
            args[0](*arg[1:])
    else:
        with Pool() as p:
            p.starmap(arg[0], arg[1:])


def estimate_ieeg_prf(ieeg_file, method, freq=(60, 80)):
    with ieeg_file.open('rb') as f:
        data = load(f)

    stimuli = data.attr['stimuli']

    data = select(data, freq=freq)
    data = math(data, operator_name='mean', axis='time')
    data = math(data, operator_name='mean', axis='freq')
    data = concatenate(data, 'trial')

    compute_prf(ieeg_file, data.data[0], data.chan[0], stimuli, method)

    return replace_extension(ieeg_file, 'prf.tsv')


def estimate_bold_prf(bold_file, method):
    prf_task = Task(bold_file)
    stimuli = read_prf_stimuli(prf_task)

    region_idx = 11143
    aparc = nload(str(data_aparc))

    img = nload(str(prf_task.filename))
    mri = img.get_data()

    roi = select_region(aparc, lambda x: x == region_idx)
    roi_idx = apply_affine(img.affine, array(where(roi.get_data() > 0)).T)
    roi_str = [f'{xyz[0]:.2f},{xyz[1]:.2f},{xyz[2]:.2f}' for xyz in roi_idx]

    dat = mri[roi.get_data() > 0, :]

    output = compute_prf(bold_file, dat, roi_str, stimuli, method)

    images = ['X', 'Y', 'SIGMA', 'BETA']
    for i in range(len(output)):
        nii_file = replace_extension(bold_file, 'prf' + images[i] + '.nii.gz')

        out = zeros(mri.shape[:3])
        out[roi.get_data() > 0] = output[i]
        x_img = Nifti1Image(out, img.affine)
        x_img.to_filename(str(nii_file))


def compute_prf(input_file, dat, indices, stimuli, method):

    tsv_file = replace_extension(input_file, 'prf.tsv')

    x = []
    y = []
    sigma = []
    beta = []

    with tsv_file.open('w') as f:
        f.write(f'channel\tx\ty\tsigma\tbeta\n')
        for i, index in enumerate(indices):
            if method == 'analyzePRF':
                output = fit_analyzePRF(stimuli, dat[i, :])
                f.write(f'{index}\t{output.x[0]}\t{output.x[1]}\t{output.x[2]}\t{output.x[3]}\n')
                results = output.x

            elif method == 'popeye':
                output = fit_popeye(stimuli, dat[i, :])
                f.write(f'{index}\t{output.estimate[0]}\t{output.estimate[1]}\t{output.estimate[2]}\t{output.estimate[3]}\n')
                results = output.estimate

            x.append(results[0])
            y.append(results[1])
            sigma.append(results[2])
            beta.append(results[3])

            f.flush()

    return array(x), array(y), array(sigma), array(beta)



from boavus.workflow.corr_ieeg_fmri import workflow_corr_ieeg_fmri
from nipype import config

from .paths import (ANALYSIS_PATH,
                    BIDS_PATH,
                    FREESURFER_PATH,
                    task_anat,
                    task_fmri,
                    task_ieeg,
                    elec,
                    parameters,
                    )


LOG_PATH = ANALYSIS_PATH / 'log'
config.update_config({
    'logging': {
        'log_directory': LOG_PATH,
        'log_to_file': True,
        },
    'execution': {
        'crashdump_dir': LOG_PATH,
        'keep_inputs': 'false',
        'remove_unnecessary_outputs': 'false',
        },
    })


def test_workflow_fmri():

    w = workflow_corr_ieeg_fmri(parameters, FREESURFER_PATH)
    w.base_dir = str(ANALYSIS_PATH)

    node = w.get_node('input')
    node.inputs.subject = 'sub-delft'
    node.inputs.T1w = str(task_anat.get_filename(BIDS_PATH))
    node.inputs.bold = str(task_fmri.get_filename(BIDS_PATH))
    node.inputs.electrodes = str(elec.get_filename(BIDS_PATH))
    node.inputs.ieeg = str(task_ieeg.get_filename(BIDS_PATH))

    w.run()



from bidso import file_Core
from numpy import gradient, argmin, argmax
from shutil import copy
from bidso.utils import read_tsv


def collect_corr(in_files, ecog_files, fmri_files, output_dir):

    rsquared_dir = output_dir / 'rsquared'
    rsquared_dir.mkdir(parents=True, exist_ok=True)

    ecog_dir = output_dir / 'ecog'
    ecog_dir.mkdir(parents=True, exist_ok=True)

    fmri_dir = output_dir / 'fmri'
    fmri_dir.mkdir(parents=True, exist_ok=True)

    for in_file in in_files:
        copy(in_file, rsquared_dir)

    for in_file in ecog_files:
        copy(in_file, ecog_dir)

    for in_file in fmri_files:
        copy(in_file, fmri_dir)

    _compute_summary(in_files, output_dir)


def _compute_summary(in_files, output_dir):

    summary_file = output_dir / 'summary_per_subject.tsv'

    with summary_file.open('w') as f:

        f.write(f'subject\tsession\ttask\tacquisition\tsize_at_peak\tr2_at_peak\tslope_at_peak\tintercept_at_peak\tsize_at_concave\tr2_at_concave\tdiff_r2\n')

        for corr_file in in_files:

            corr_tsv = read_tsv(corr_file)
            size_max, r2_max, slope, intercept = corr_tsv[argmax(corr_tsv['Rsquared'])]

            deriv = gradient(gradient(corr_tsv['Rsquared']))
            size_deriv, r2_deriv, *dummy = corr_tsv[argmin(deriv)]

            file_info = file_Core(corr_file)

            f.write(f'{file_info.subject}\t{file_info.session}\t{file_info.task}\t{file_info.acquisition}\t{size_max}\t{r2_max}\t{slope}\t{intercept}\t{size_deriv}\t{r2_deriv}\t{r2_max - r2_deriv}\n')
